import logging
import inspect
from tornado.web import RequestHandler
from etornado.error_code_manager import HTTP_OK, ErrorCode, error_code_manager


class AggregateHttMethodMeta(type):

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls.head = cls.process
        cls.get = cls.process
        cls.post = cls.process
        cls.delete = cls.process
        cls.patch = cls.process
        cls.put = cls.process
        cls.options = cls.process


class BaseHandler(RequestHandler, metaclass=AggregateHttMethodMeta):

    def initialize(self, thread_pool_executor):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.thread_pool_executor = thread_pool_executor
        self.url_args = []

    def write_error(self, status_code, **kwargs):
        error_code = kwargs.get("ec")
        error_info = error_code_manager.format_error_info(error_code, **kwargs)
        self.logger.error(
            "write error for url[%s], method[%s], error_info[%s]",
            self.request.uri, self.request.method, error_info)
        self.write(error_info)
        self.finish()

    async def process(self, *args, **kwargs):
        method = getattr(self, "do_" + self.request.method.lower(), None)
        self.url_args = args
        if method is None:
            self.send_error(HTTP_OK,
                            ec=ErrorCode.UNSUPPORTED_METHOD,
                            method=self.request.method,
                            url=self.request.uri)
            return
        try:
            r = method()
            if inspect.isawaitable(r):
                r = await r
            error_code, result, error_args = self.__parse_result(r)
        except Exception as e:
            self.logger.exception("exception [%s] catched when call [%s]", e, method)
            raise e
        response = error_code_manager.format_error_info(error_code, **error_args)
        if error_code == ErrorCode.NONE:
            response["result"] = result
        self.write(response)
        self.finish()

    def __parse_result(self, r):
        if isinstance(r, tuple):
            if len(r) != 2:
                raise Exception("tuple result [%s]'s length must be 2")
            if r[0] != ErrorCode.NONE:
                if not isinstance(r[1], dict):
                    raise Exception("error info args must be dict")
                return r[0], {}, r[1]
            else:
                return r[0], r[1], {}
        return ErrorCode.NONE, r, {}



from etornado.buildin_handlers.base_handler import BaseHandler
from etornado.error_code_manager import HTTP_OK, ErrorCode


class UnsupportedUrlHandler(BaseHandler):

    def process(self, *args, **kwargs):
        self.send_error(HTTP_OK, ec=ErrorCode.UNSUPPORTED_URL, url=self.request.uri)



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()
 
setuptools.setup(
    name="etornado",
    version="0.0.3",
    author="sheng_xc",
    author_email="sheng_xc@126.com",
    description="a wrapper for tornado",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    install_requires=[
        "tornado==6.0.3"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7'
)



import tornado.ioloop
import tornado.web
import tornado.httpserver
import logging.config
from concurrent.futures import ThreadPoolExecutor
from etornado.buildin_handlers.unsupported_url_handler import UnsupportedUrlHandler


class Application(object):
    def __init__(self, port, process_count=0, thread_pool_size=0, logging_config=None):
        if logging_config is not None:
            logging.config.fileConfig(logging_config)
        self.logger = logging.getLogger(self.__class__.__name__)
        self.stopping = False
        self.handlers = []
        self.port = port
        self.process_count = process_count
        self.thread_pool_executor = None
        if thread_pool_size > 0:
            self.thread_pool_executor = ThreadPoolExecutor(
                    max_workers=thread_pool_size)

    def register_handler(self, url, klass):
        self.handlers.append((url, klass))

    def stop(self):
        self.logger.info("call stop")
        self.stopping = True

    def run(self):
        self.logger.info("start service ...")
        handlers = self.get_handlers()
        self.logger.info("register %d handlers[%s]", len(handlers), handlers)
        app = tornado.web.Application(handlers)
        server = tornado.httpserver.HTTPServer(app)
        server.bind(self.port)
        server.start(self.process_count)
        tornado.ioloop.PeriodicCallback(self.check_stop, 100).start()
        tornado.ioloop.IOLoop.current().start()

    def get_handlers(self):
        result = []
        handler_args = dict(thread_pool_executor=self.thread_pool_executor)
        for url, handler in self.handlers:
            result.append((url, handler, handler_args))
        result.append((r".*", UnsupportedUrlHandler, handler_args))
        return result

    def check_stop(self):
        if self.stopping:
            self.logger.info("stop service ...")
            if self.thread_pool_executor:
                self.thread_pool_executor.shutdown(False)
            tornado.ioloop.IOLoop.current().stop()
            self.logger.info("stop service success!!!")



from enum import Enum, unique
import logging
from etornado.singleton import SingletonMeta

HTTP_OK = 200

@unique
class ErrorCode(Enum):
    NONE = 0
    UNKNOWN = 100
    UNSUPPORTED_URL = 101
    UNSUPPORTED_METHOD = 102


ERROR_INFO_MAP = {
    ErrorCode.NONE: "ok",
    ErrorCode.UNKNOWN: "unknown error",
    ErrorCode.UNSUPPORTED_URL: "unsupported url [{url}]",
    ErrorCode.UNSUPPORTED_METHOD:
    "unsupported method [{method}] for url [{url}]",
}


class ErrorCodeManager(object, metaclass=SingletonMeta):

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.error_infos = {}
        self.register_error_enum(ErrorCode, ERROR_INFO_MAP)

    def register_error_enum(self, error_enum, error_info_map):
        for name, member in error_enum.__members__.items():
            error_msg = error_info_map.get(member, "")
            if not error_msg:
                self.logger.warn("error msg of [%s] is empty", member)
            self.register_error_code(member.value, error_msg)

    def register_error_code(self, error_num, error_msg):
        if error_num in self.error_infos:
            self.logger.error("error_num [%d] has be registered"
                    " with message[%s]", error_num, error_msg)
            return False
        self.error_infos[error_num] = error_msg

    def format_error_info(self, error_code, **kwargs):
        if error_code is None:
            error_code = ErrorCode.UNKNOWN
        if isinstance(error_code, Enum):
            error_code = error_code.value
        error_message = self.error_infos.get(error_code, "").format(**kwargs)
        return {"error_code": error_code, "error_message": error_message}


error_code_manager = ErrorCodeManager()
